﻿#pragma TextEncoding="UTF-8"
#pragma rtGlobals=3
#pragma IgorVersion=8
#pragma IndependentModule=TextEditTools
#pragma version=1.6

// A compendium of text-editing tools

// Contributors:
// https://www.wavemetrics.com/user/tony
// https://www.wavemetrics.com/user/chozo

// To format selection as a wrapped comment: ctrl-4/cmd-4
// To prettify selection: ctrl-5 (Windows) or cmd-5 (Mac)
// To align comments in the selected code use:
// Left-aligned: cmd-7 (Mac), ctrl-7 (Windows)
// Right-aligned: cmd-8 (Mac), ctrl-8 (Windows)

// For comment aligning, dial in your procedure window settings here:
// You can find the default tab size under Procedure->Document Settings...
Strconstant FontName = "Courier New"
Constant FontSize = 11
Constant DefaultTab = 20	// in points

// Set line length for word wrappping comments here:
Constant kCommentWidth=72 // word wrapping is done as a 'greedy' wrap

// Customize prettification here:
Strconstant ksIgnoreList="" // function names, operation names and keywords to ignore
Strconstant ksCustomList="" // add or override items
Constant kMinWordLength=2

// Edit menu shortcuts here:
Menu "Edit"
	"Comment Wrap Clipboard", /Q, WrapScrap()
	"Unwrap Clipboard", /Q, UnwrapScrap()
	"Prettify Clipboard Code", /Q, PrettifyScrap()
	"Wrap Selected Comment/4", /Q, WrapSelection()
	"Prettify Selected Code/5", /Q, PrettifySelection()
	"Align Comments in Clipboard Code (Right)", /Q, PutScrapText AlignComments(GetScrapText())
	"Align Comments in Selected Code (Left)/7", /Q, AlignCommentsInSelection(1)
	"Align Comments in Selected Code (Right)/8", /Q, AlignCommentsInSelection(0)
End

Function WrapScrap()
	PutScrapText CommentWrap(GetScrapText(), kCommentWidth)
End

Function UnwrapScrap()
	PutScrapText CommentUnwrap(GetScrapText())
End

Function PrettifyScrap()
	PutScrapText Prettify(GetScrapText())
End

Function WrapSelection()
	
	String strScrap = GetScrapText()
	PutScrapText ""
	DoIgorMenu "Edit" "Copy"
	WrapScrap()
	DoIgorMenu "Edit" "Paste"
	PutScrapText strScrap // leave clipboard state unchanged
End

// returns string wrapped at V_wrap characters, commented,
// and indented to match indent at start of s_in.
Function /T CommentWrap(String s_in, Variable v_wrap)
	
	if(strlen(s_in)==0)
		return ""
	endif
	String s_out="", s_char=""
	s_in=ReplaceString("\t", s_in, "   ")
	
	// try to guess how far we want to indent comment
	Variable indent=0, maxIndent=25
	for(indent=0;indent<maxIndent;indent+=1)
		if(cmpstr(s_in[indent], " "))
			break
		endif
	endfor
	
	s_in=ReplaceString("//", s_in, " ")
	s_in=ReplaceString("\r", s_in, " ")
	
	// condense whitespace
	Variable len
	do
		len=strlen(s_in)
		s_in=ReplaceString("  ", s_in, " ")
	while(strlen(s_in)!=len)
	
	Variable i=0,j=0,k=0
	// i is character count in s_in
	// j is character count in current line of s_out
	do
		s_char=s_in[i]
		i+=1
		if(j==0)
			for(k=0;k<indent;k+=1,j+=1)
				s_out+=" "
			endfor
			s_out+="// "
			j+=3
			if (cmpstr(s_char, " "))
				s_out+=s_char
				j+=1
			endif
		elseif (j>v_wrap)
			if(cmpstr(s_char, " "))
				do
					j-=1
					if (cmpstr(s_out[strlen(s_out)-(v_wrap+1-j)], " ")==0)
						break
					endif
				while(j>indent+3)
				if(j>indent+3)
					s_out=s_out[0,strlen(s_out)-(v_wrap+1-j)]
					i-=(v_wrap+1-j)
				endif
			endif
			s_out+="\r"
			j=0
		else
			s_out+=s_char
			j+=1
		endif
	while(i<len)
	s_out=ReplaceString("   ", s_out, "\t")
	
	return s_out
End

// returns unwrapped string with comment indicators removed
Function /T CommentUnwrap(String s)
	
	s=ReplaceString("\t", s, "   ")
	s=ReplaceString("//", s, "")
	s=ReplaceString("\r\r", s, "<$$> ")
	s=ReplaceString("\r", s, " ")
	s=ReplaceString("<$$>", s, "\r\r")
	
	// condense whitespace
	Variable len
	do
		len=strlen(s)
		s=ReplaceString("  ", s, " ")
	while(strlen(s)!=len)
	
	do
		len=strlen(s)
		s=ReplaceString("\r ", s, "\r")
	while(strlen(s)!=len)
	
	do
		s[0,0]=SelectString(cmpstr(s[0]," "), "", s[0])
	while(strlen(s) && cmpstr(s[0]," ")==0)
	
	return s
End

// --------------------- Prettify -------------------------

Function PrettifySelection()
	
	String strScrap = GetScrapText()
	PutScrapText ""
	DoIgorMenu "Edit" "Copy"
	PutScrapText Prettify(GetScrapText())
	DoIgorMenu "Edit" "Paste"
	PutScrapText strScrap // leave clipboard state unchanged
End

// Returns a string with known function and operation names case-corrected.
// Commented and quoted text is ignored. Trailing whitespace is eradicated.
// Blocks of code representing proc pictures are ignored, provided that the
// selection starts before the Picture keyword.
Function /S Prettify(String strText)
	
	if(strlen(strText)==0)
		return ""
	endif
	
	String strList = FunctionList("*",";","") + OperationList("*",";","")
	String keyWords="End;EndMacro;EndStructure;Function;Macro;Picture;Proc;Structure;Window;"
	keyWords+="Constant;DoPrompt;GalleryGlobal;hide;IgorVersion;IndependentModule;Menu;ModuleName;MultiThread;"
	keyWords+="Override;Popup;ProcGlobal;Prompt;root;rtGlobals;Static;Strconstant;Submenu;TextEncoding;ThreadSafe;version;"
	String subtypes="ButtonControl;CameraWindow;CDFFunc;CheckboxControl;CursorStyle;DrawUserShape;FitFunc;"
	subtypes+="GizmoPlot;Graph;GraphMarquee;GraphStyle;GridStyle;Layout;LayoutMarquee;LayoutStyle;ListBoxControl;"
	subtypes+="Panel;PopupMenuControl;SetVariableControl;SliderControl;TabControl;Table;TableStyle;"
	String objectRefs="DFREF;FUNCREF;NVAR;STRUCT;SVAR;WAVE;"
	String flowControl="AbortOnRTE;AbortOnValue;break;case;catch;continue;default;do;else;elseif;endfor;endif;"
	flowControl+="endswitch;endtry;for;if;return;strswitch;switch;try;while;"
	strList+=keyWords+subtypes+objectRefs+flowControl

	// remove words in ksIgnoreList, substitute or add words in ksCustomList
	strList = RemoveFromList(ksIgnoreList+ksCustomList, strList, ";", 0) + ksCustomList
	WAVE /T wList = ListToTextWave(strList, ";")
	// remove words with fewer than kMinWordLength characters
	Variable i
	for(i=numpnts(wList)-1; i>=0; i-=1)
		if(strlen(wList[i]) < kMinWordLength)
			DeletePoints i, 1, wList
		endif
	endfor
	
	Variable endsWithReturn = (cmpstr(strText[strlen(strText)-1], "\r") == 0)
	WAVE /T wText = ListToTextWave(strText, "\r")
	Variable lastLine = numpnts(wText) - 1
	wText += SelectString(p<(lastLine) || endsWithReturn, "", "\r")
	
	// create a mask for lines that should not be altered
	Variable numLines=numpnts(wText)
	Make /free/N=(numLines) wIgnore=0
	Grep /Q/INDX/E="(?i)^\s*Picture\b" wText
	WAVE W_Index
	Variable vLine
	
	for (i=0;i<numpnts(W_Index);i+=1)
		vLine=w_index[i]+1
		if(vLine>numLines)
			break
		endif
		do
			wIgnore[vLine]=1
			vLine+=1
		while(vLine<numLines && GrepString(wText[vLine],"(?i)^\s*End\b")==0)
	endfor

	multithread wText = PrettifyLine(wText, wList, wIgnore)
	wfprintf strText, "%s", wText
	return strText
End

// strLine should include \r if one was present in input text
ThreadSafe Function /S PrettifyLine(String strLine, WAVE /T wordList, Variable ignore)
	
	if(ignore)
		return strLine
	endif
	
	WAVE /B mask = QuoteAndCommentMask(strLine)
	String strCode="", word=""
	int bytes = -1, len=numpnts(mask)
	// find position of last unmasked character in line
	Duplicate /free mask rmask
	Reverse mask /D=rmask
	FindValue /I=0 rmask // find first 0 in reversed mask wave
	if(v_value==-1) // entire line is masked
		strCode=""
	else
		strCode=strLine[0,len-1-v_value]
	endif
	// strCode is stripped of commented text

	do // loop through words in line
		SplitString /E="([[:alnum:]]+)(.*)" strCode, word, strCode
		FindValue /TEXT=word/TXOP=4/Z wordList
		if(V_value > -1)
			strLine = ReplaceWord(word, strLine, wordList[V_value], mask)
		endif
	while(strlen(strCode))

	// remove trailing whitespace, ie. spaces and tabs that follow non-whitespace at end of line
	if(GrepString(strLine, "\S[[:blank:]]+\r$"))
		do
			bytes = strlen(strLine)
			strLine = ReplaceString(" \r", strLine, "\r")
			strLine = ReplaceString("\t\r", strLine, "\r")
		while(bytes > strlen(strLine))
	endif
	
	return strLine
End

ThreadSafe Function /WAVE QuoteAndCommentMask(String strLine)
		
	Variable startByte, endByte, len
	len=strlen(strLine)
	Make /B/free/N=(len) mask=0
	startByte=0
	do
		startByte=strsearch(strLine, "\"", startByte)
		if(startByte==-1) // no more quoted text before end of strLine
			break
		endif
		endByte=startByte
		do
			endByte=strsearch(strLine, "\"", endByte+1)
			if(endByte==-1)
				endByte=len-1
				break
			endif
			// endByte is possible end of quote
			// remove escaped backslashes!
			strLine[startByte,endByte]=ReplaceString("\\\\", strLine[startByte,endByte], "  ")
		while(cmpstr(strLine[endByte-1], "\\")==0) // ignore escaped quotes
		// found end of quote
		mask[startByte, endByte]=1
		startByte=endByte+1
	while(1) // look for next quoted text
	// quoted text is now masked
	startByte=0
	do
		startByte=strsearch(strLine, "//", startByte)
		if(startByte==-1) // no comment before end of strLine, we're done
			return mask
		endif
		if(mask[startByte]==1)
			startByte+=1
			continue
		endif
		// startByte is start of comment
		mask[startByte,len-1]=1
		return mask
	while(1)
End

// Wrapper for ReplaceString, replaces whole words
// characters inStr[i] where mask[i]==1 are ignored
ThreadSafe Function /S ReplaceWord(replaceThisWord, inStr, withThisWord, mask)
	String replaceThisWord, inStr, withThisWord
	WAVE /B mask
	
	String substring=""
	Variable startByte=0, endByte=0
	
	if(strlen(replaceThisWord)==0)
		return inStr
	endif

	do // loop through matches
		startByte = strsearch(inStr, replaceThisWord, endByte, 2)
		if(startByte == -1)
			return inStr
		endif
		endByte = startByte + strlen(replaceThisWord)
		if(mask[startbyte]==1)
			continue
		endif
		substring = inStr[startByte-1, endByte+1] // will be clipped to 0, strlen(inStr)
		// check that match is whole word
		if(GrepString(substring, "(?i)\\b" + replaceThisWord + "\\b"))
			substring = ReplaceString(replaceThisWord, substring, withThisWord)
		endif
		inStr = inStr[0, startByte-2] + substring + inStr[endByte+2, Inf]
	while(1)
End

// ---------------- transpose paste --------------------------

Menu "Edit", hideable, dynamic
	SelectString((WinType("")==2), "", "Transpose Paste"), /Q, TextEditTools#TransposePaste()
End

Function TransposePaste()
	
	String clip=GetScrapText(), strOut=""
	Variable rows=ItemsInList(clip, "\r")
	Variable columns=ItemsInList(StringFromList(0,clip,"\r"), "\t")
	Variable i,j
	for(i=0;i<columns;i+=1)
		for(j=0;j<rows;j+=1)
			strOut+=StringFromList(i,StringFromList(j, clip,"\r"), "\t")+"\t"
		endfor
		strOut=RemoveEnding(strOut,"\t")
		strOut+="\r"
	endfor
	PutScrapText strOut
	DoIgorMenu "Edit", "Paste"
	PutScrapText clip // leave state of clipboard unchanged
End

// --------------------- align comments -------------------------

Function AlignCommentsInSelection(Variable LeftOrRight)
	String strScrap = GetScrapText()
	DoIgorMenu "Edit" "Copy"
	PutScrapText AlignComments(GetScrapText(),AligntoLeft=LeftOrRight)
	DoIgorMenu "Edit" "Paste"
	PutScrapText strScrap // leave clipboard state unchanged
End

// For setting a determined tab position of all selected comments.
// Execute from the commandline as AlignComments#toTabPos(desiredPos).
Function toTabPos(Variable TabPos)															// for aligning all comments to a certain tab position
	DisplayProcedure/W=$StringFromList(0,WinList("*", ";","WIN:128"))				// bring to front
	String strScrap = GetScrapText()
	DoIgorMenu "Edit" "Copy"
	PutScrapText AlignComments(GetScrapText(),TabPos=TabPos)
	DoIgorMenu "Edit" "Paste"
	PutScrapText strScrap
End

// The optional variable TabPos lets you choose a fixed offset (in tab widths) for all comments.
// For example, TabPos = 30 will position all comments at 30 tabs, if possible.
Function/S AlignComments(String strText, [Variable TabPos, Variable AlignToLeft])
	Variable LorR = 0
	if(!ParamIsDefault(AlignToLeft))
		LorR = AlignToLeft
	endif
	
	Variable fsPix = FontSize * ScreenResolution/72										// font size in pix
	Variable TabPix = floor(DefaultTab * ScreenResolution/72)							// tab width in pix
	
	WAVE/T wText = ListToTextWave(strText, "\r")
	Variable endsWithReturn = (cmpstr(strText[strlen(strText)-1], "\r") == 0)
	Variable entries = numpnts(wText)

	String strEverything = "", strStartOfFirst	 = "", strEndOfFirst = wText[0]		// extract the start of the first line from the whole text
	strEverything = ProcedureText("",0,StringFromList(0,WinList("*", ";","WIN:128")))
	strEverything = ReplaceString(strText,strEverything,"\rCUT\r")					// create a break at selected text
	strEverything = StringFromList(0, strEverything, "\rCUT\r")						// only the stuff before the selection
	if (cmpstr(strEverything[strlen(strEverything)-1], "\r") != 0 && strlen(strEverything)>0)	// only if it is not a full line
		WAVE/T wBegin = ListToTextWave(strEverything, "\r")
		strStartOfFirst	= wBegin[numpnts(wBegin)-1]										// the extra part of the first line
	endif

	if (strlen(strStartOfFirst) > 0)															// put the full firt line in here
		wText[0] = strStartOfFirst + strEndOfFirst											// this is needed to calculate the correct line length
	endif
	wText += SelectString(p<(entries-1) || endsWithReturn, "", "\r")

	Make/D/FREE/N=(entries) CodeSize = 0													// saves the code size in 'tab widths'
	Make/T/FREE/N=(entries) CodeOnly, CommentOnly
	
	Variable i,j, bytes = 0, strSize = 0
	String strLine = "", strCode = ""
	
	for (i = 0; i < entries; i += 1)
		strLine = wText[i]
		//bytes = strsearch(strLine, "//",Inf,1)
		bytes = startOfComment(strLine)														// is there a comment?
		bytes = bytes == strlen(strLine) ? -1 : bytes										// no comment found

		if (i == 0 && strlen(strStartOfFirst) > bytes)									// if the full comment in not selected -> skip
			bytes = -1
		endif
		
		if(bytes > 1)
			strCode = strLine[0,bytes-1]
			if(GrepString(strCode, "\S")) 													// line contains non-whitespace
				CodeOnly[i]		= strCode														// just the code
				CommentOnly[i]	= strLine[bytes,Inf]										// just the comment
				if (cmpstr(strCode[strlen(strCode)-1], " ") == 0)
					CodeOnly[i] 	= RemoveEnding(CodeOnly[i] , " ")						// get rid of a possible space character in front of the comments
					CodeOnly[i] 	+= "\t"
				endif
				
				// Below code calculates the line length in units of 'tab length'.
				// Multiplication by TabPix would give the length in pix.
				WAVE/T wTabList = ListToTextWave(CodeOnly[i], "\t")						// split code line by tabs
				for (j = 0; j < DimSize(wTabList,0); j += 1)
					strCode = wTabList[j]
					if (strlen(strCode) > 0)													// tab only or code?
						strSize = FontSizeStringWidth(FontName,fsPix,0,strCode,"")		// get the length of all characters
						if (mod(strSize,TabPix) < 0.05)										// does this have roughly the size of a certain no. of tabs?
							strSize = ceil(strSize/TabPix + 1)								// a full tab length is added
						else
							strSize = ceil(strSize/TabPix)									// this will add a fractional tab length
						endif
						CodeSize[i] += strSize
					else
						CodeSize[i] += 1															// if it's just a tab then add one tab length
					endif
				endfor
				if (cmpstr(strCode[strlen(strCode)-1], "\t") == 0)						// no tab in last entry -> not aligned to the next tab position
					CodeSize[i] -= 1
				endif
			endif
		endif
	endfor
	
	wText[0] = strEndOfFirst + "\r"															// put only the end line back (in case there is no modification)
	if (strlen(strStartOfFirst) > 0)															// remove the first part of the first line again
		CodeOnly[0] = ReplaceString(strStartOfFirst,CodeOnly[0],"")
	endif
	
	Variable AlignDistance = 0																// how many tabs define the new alignment
	if(!ParamIsDefault(TabPos))
		AlignDistance = TabPos
	else
		Extract/Free CodeSize,SizeVals,CodeSize!=0
		AlignDistance = LorR == 1 ? max(WaveMin(SizeVals),AlignDistance) :  max(WaveMax(SizeVals),AlignDistance)
	endif
	
	CodeSize = CodeSize[p] > 0 ? AlignDistance - CodeSize[p] : 0						// how many tabs to add or subtract
	for (i = 0; i < entries; i += 1)
		String newTabs = ""
		if(abs(CodeSize[i]) > 0)
			for (j = 0; j < abs(CodeSize[i]); j += 1)
				if (CodeSize[i] < 0)
					CodeOnly[i] = RemoveEnding(CodeOnly[i],"\t")
				else
					newTabs += "\t"
				endif
			endfor
			wText[i] = CodeOnly[i] + newTabs + CommentOnly[i]
		endif
	endfor
		
	wfprintf strText, "%s", wText
	return strText
End

// comment search function by Tony Withers
Function startOfComment(String strLine)
    Variable startByte, commentByte, startQuote, endQuote, lineLength
    lineLength=strlen(strLine)
    startByte=0
    do
        if(startByte>=(lineLength-1))
            return lineLength
        endif
        commentByte = strsearch(strLine, "//", startByte)
        if(commentByte == -1)
            return lineLength
        endif
        // found "//"
        startQuote=strsearch(strLine, "\"", startByte)
        if(startQuote==-1 || startQuote>commentByte) // no quotes before //, we're done
            return commentByte
        endif
        // we have a quotation mark before //
       
        do
            endQuote=startQuote
            do
                endQuote=strsearch(strLine, "\"", endQuote+1)
                if(endQuote==-1)
                    return lineLength
                endif
                // endQuote is possible end of quote
                // remove escaped backslashes!
                strLine[startQuote,endQuote]=ReplaceString("\\\\", strLine[startQuote,endQuote], "  ")
            while(cmpstr(strLine[endQuote-1], "\\")==0) // ignore escaped quotes
           
            // found end of quote
            if(endQuote>commentByte) // commentByte was within commented text
                startByte=endQuote+1
                break // look for another comment mark
            else
                startQuote=strsearch(strLine, "\"", endQuote+1)
                if(startQuote==-1 || startQuote>commentByte) // no quotes before //, we're done
                    return commentByte
                endif
            endif
            // if we get to here we've found another start-of-quoted-text before //
        while(1) // next quoted text before //
    while(1) // next comment marker
End